/*
 * This file is provided under a Simplified BSD License.
 *
 * Copyright (C) 2017 Atmark Techno, Inc. All Rights Reserved.
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
 * SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
 * OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <signal.h>
#include <termios.h>
#include <unistd.h>
#include <libgen.h>

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <errno.h>
#include <time.h>

#include "crc8.h"

#define CRC8_POLYNOMIAL (0x85)      /* x^8 + x^7 + x^2 + 1 */
DECLARE_CRC8_TABLE(crc8_table);

#define HDR_CMD (0xAA)

#define __unused __attribute__((unused))
#define ARRAY_SIZE(a) (sizeof(a) / sizeof((a)[0]))
#define SERIAL_BAUDRATE B115200
#define BUF_SIZE (256)
#define HDR_CNT_CRC_LENGTH	(4)
#define RTC_TMR_SND_CMD		(0x01)
#define SRT_MSR_SND_CMD		(0x02)
#define STP_MSR_SND_CMD		(0x03)
#define RTM_MSR_SND_CMD		(0x08)
#define RTC_TMR_SND_LENGTH	(8 + HDR_CNT_CRC_LENGTH)
#define SRT_MSR_SND_LENGTH	(2 + HDR_CNT_CRC_LENGTH)
#define STP_MSR_SND_LENGTH	(1 + HDR_CNT_CRC_LENGTH)
#define RTM_MSR_SND_LENGTH	(1 + HDR_CNT_CRC_LENGTH)
#define RTC_TMR_RCV_LENGTH	(2 + HDR_CNT_CRC_LENGTH)
#define SRT_MSR_RCV_LENGTH	(2 + HDR_CNT_CRC_LENGTH)
#define STP_MSR_RCV_LENGTH	(2 + HDR_CNT_CRC_LENGTH)
#define RTC_MSR_RCV_LENGTH	(17 + HDR_CNT_CRC_LENGTH)

#define DATA(a,b,c) ((a << 16) | (b << 8) | c)

#define VOLTAGE_STEP	1	/* mV */
#define TO_V(v)		((float)v * VOLTAGE_STEP / 1000)	/* V */
#define CURRENT_STEP	1 / 128	/* mA */
#define TO_MA(v)	((float)v * CURRENT_STEP)		/* mA */
#define POWER_STEP	5	/* mW */
#define TO_W(v)		((float)v * POWER_STEP / 1000)		/* W */


#define REQUEST_INTERVAL	(2)	/* 秒 */

static int terminated = 0;		/* 終了シグナル発生フラグ */

/**
 * シグナルハンドラ
 * @param signum シグナル番号(不使用)
 */
static void sig_handler(__unused int signum)
{
	/* 終了シグナル発生を記録 */
	terminated = 1;
}

/**
 * シグナルハンドラ設定関数
 */
static int install_sig_handler(void)
{
	int term_sigs[] = { /* 終了シグナル種類 */
		SIGHUP, SIGINT, SIGQUIT, SIGPIPE, SIGTERM
	};
	struct sigaction sa;
	unsigned int i;
	int ret;

	/* ハンドラ関数を設定 */
	memset(&sa, 0, sizeof(sa));
	sa.sa_handler = sig_handler;

	/* 各シグナルに対して関連付け */
	for(i = 0; i < ARRAY_SIZE(term_sigs); i++) {
		ret = sigaction(term_sigs[i], &sa, NULL);
		if (ret < 0) {
			perror("sigaction");
			return -1;
		}
	}

	return 0;
}

/**
 * シリアルポート設定復元関数
 */
static void restore_serial(int fd, struct termios *old_tio)
{
	int ret;

	/* シリアルポートの設定を元に戻す */
	ret = tcsetattr(fd, TCSANOW, old_tio);
	if (ret < 0)
		perror("tcsetattr");
}

/**
 * シリアルポート設定関数
 * @param fd 設定するシリアルポートファイルディスクリプタ
 */
static int setup_serial(int fd, struct termios *old_tio)
{
	struct termios tio;
	int ret;

	/* 現在のシリアルポートの設定を退避する */
	ret = tcgetattr(fd, old_tio);
	if (ret) {
		perror("tcgetattr");
		return -1;
	}

	/* 新しいシリアルポートの設定 */
	memset(&tio, 0, sizeof(tio));

	/* ブレーク文字無視/パリティなし */
	tio.c_iflag = IGNBRK | IGNPAR;
	/* フロー制御なし/8bit/非モデム/受信可 */
	tio.c_cflag = CS8 | CLOCAL | CREAD;
	/* キャラクタ間タイマー無効 */
	tio.c_cc[VTIME] = 0;
	/* 最低1文字送信/受信するまでブロックする */
	tio.c_cc[VMIN] = 1;
	/* 入出力ボーレート */
	ret = cfsetspeed(&tio, SERIAL_BAUDRATE);
	if (ret < 0) {
		perror("cfsetspeed");
		return -1;
	}

	/* バッファ内のデータをフラッシュ */
	ret = tcflush(fd, TCIFLUSH);
	if (ret < 0) {
		perror("tcflush");
		return -1;
	}

	/* 新しいシリアルポート設定を適用 */
	ret = tcsetattr(fd, TCSANOW, &tio);
	if (ret < 0) {
		perror("tcsetattr");
		return -1;
	}

	return 0;
}


/**
 * RTC タイマー設定コマンド作成関数
 * @param fyload 作成ペイロード格納用変数
 */
static void create_timer_payload(uint8_t *payload)
{
	time_t timer;
	struct tm *local;
	int year;

	/* 現在時刻を取得 */
	timer = time(NULL);
	local = localtime(&timer); /* 地方時に変換 */
	year = local->tm_year + 1900;
	year %= 100;

	/* 現在時刻を設定 */
	payload[0] = RTC_TMR_SND_CMD;
	payload[1] = local->tm_sec;
	payload[2] = local->tm_min;
	payload[3] = local->tm_hour;
	payload[4] = local->tm_mday;
	payload[5] = (local->tm_mon + 1);
	payload[6] = year;
	payload[7] = local->tm_wday;
}

/**
 * コマンド生成関数
 * @param cmd 作成コマンド格納用変数
 * @param payload ペイロード
 * @param payload_size ペイロード数
 */
static void create_command(uint8_t *cmd, uint8_t *pld, size_t pld_size)
{
	size_t cmd_size = pld_size - HDR_CNT_CRC_LENGTH;

	/* ヘッダーを設定 */
	cmd[0] = HDR_CMD;	/* ヘッダー指定(固定値) */

	/* カウントを設定 */
	cmd[1] = (uint8_t)cmd_size;			/* LoByte */
	cmd[2] = (uint8_t)((cmd_size >> 8) & 0xFF);	/* HiByte */

	/* ペイロードの値を設定 */
	memcpy(&cmd[3], pld, cmd_size);

	/* CRCの値を設定 */
	cmd[cmd_size + 3] = crc8(crc8_table, pld, cmd_size, 0);
}

/**
 * write関数のwrapper
 */
static ssize_t xwrite(int fd, void *buf, size_t count)
{
	size_t len;
	ssize_t ret = -1;

	for (len = 0; len < count; len += ret) {
		ret = write(fd, buf + len, count - len);
		if (ret < 0) {
			if (errno == EINTR && !terminated) {
				ret = 0;
				continue;
			}
			perror("write");
			return ret;
		}
	}

	return ret;
}

/**
 * read関数のwrappwer
 */
static ssize_t xread(int fd, void *buf, size_t count)
{
	size_t len;
	ssize_t ret = -1;

	for (len = 0; len < count; len += ret) {
		ret = read(fd, buf + len, count - len);
		if (ret < 0) {
			if (errno == EINTR && !terminated) {
				ret = 0;
				continue;
			}
			perror("read");
			return ret;
		}
	}

	return ret;
}

/**
 * 計測データ詳細表示関数
 * @param buf 表示データ(リアルタイム計測データ転送要求コマンドの応答データ)
 */
static void disp_data_details(uint8_t *buf)
{
	int voltage, current, power;

	printf("20%02d/%2d/%2d %2d:%02d:%02d\n",
	       buf[19], buf[18], buf[17], buf[16], buf[15], buf[14]);

	current = DATA( buf[7],  buf[6],  buf[5]);
	voltage = DATA(buf[10],  buf[9],  buf[8]);
	power   = DATA(buf[13], buf[12], buf[11]);
	printf("voltage = %3.2fV , current = %4.2fmA , power = %4.2fW\n",
	       TO_V(voltage), TO_MA(current), TO_W(power));
}

/**
 * コマンド通信関数
 * @param buf 表示データ(リアルタイム計測データ転送要求コマンドの応答データ)
 */
static int communicate_command(int fd, uint8_t *wbuf, size_t wcount,
			       uint8_t *rbuf, size_t rcount)
{
        ssize_t ret;

        ret = xwrite(fd, wbuf, wcount);
        if (ret < 0)
		return -1;

        ret = xread(fd, rbuf, rcount);
        if (ret < 0)
		return -1;

        if (rbuf[4] != 0x00) {
		fprintf(stderr, "received code error\n");
		return -1;
	}

	return 0;
}

/**
 * ワットチェッカー計測開始関数
 * @param fd ファイルディスクリプタ
 */
static int start_measure(int fd)
{
	/* ペイロード:計測開始コマンド(0xFF:テスト用高速測定モード
	 *			  0x00:テスト用高速測定モード解除)
	 */
	uint8_t pld[] = {SRT_MSR_SND_CMD, 0x00,};
	uint8_t buf[BUF_SIZE];
	uint8_t cmd[SRT_MSR_SND_LENGTH];

	create_command(cmd, pld, SRT_MSR_SND_LENGTH);
	return communicate_command(fd,
				   cmd, SRT_MSR_SND_LENGTH,
				   buf, SRT_MSR_RCV_LENGTH);
}

/**
 * ワットチェッカー計測停止関数
 * @param fd ファイルディスクリプタ
 */
static int stop_measure(int fd)
{
        /* ペイロード:計測停止コマンド */
        uint8_t pld[] = {STP_MSR_SND_CMD,};
        uint8_t buf[BUF_SIZE];
        uint8_t cmd[STP_MSR_SND_LENGTH];

        create_command(cmd, pld, STP_MSR_SND_LENGTH);
        return communicate_command(fd,
				   cmd, STP_MSR_SND_LENGTH,
				   buf, STP_MSR_RCV_LENGTH);
}

/**
 * ワットチェッカー計測データ転送要求関数
 * @param fd ファイルディスクリプタ
 */
static int request_measure(int fd, uint8_t buf[BUF_SIZE])
{
	/* ペイロード:リアルタイム計測データ転送要求コマンド */
	uint8_t pld[] = {RTM_MSR_SND_CMD,};
	uint8_t cmd[RTM_MSR_SND_LENGTH];

	create_command(cmd, pld, RTM_MSR_SND_LENGTH);
	return communicate_command(fd,
				   cmd, RTM_MSR_SND_LENGTH,
				   buf, RTC_MSR_RCV_LENGTH);
}

/**
 * ワットチェッカー初期化関数
 * @param fd ファイルディスクリプタ
 */
static void init_wattchecker(int fd)
{
	/* ペイロード:RTC タイマー設定コマンド */
	uint8_t pld[RTC_TMR_SND_LENGTH];
	uint8_t buf[BUF_SIZE];
	uint8_t cmd[RTC_TMR_SND_LENGTH];

	create_timer_payload(pld);
	create_command(cmd, pld, RTC_TMR_SND_LENGTH);
	communicate_command(fd,
			    cmd, RTC_TMR_SND_LENGTH,
			    buf, RTC_TMR_RCV_LENGTH);
}

int main(int argc, char *argv[])
{
	static int fd;			/* シリアルポートファイルディスクリプタ */
	static struct termios tio;	/* 元のシリアルポート設定 */
	int ret;

	if (argc < 2) {
		printf("Usage: %s <device>\n", basename(strdup(argv[0])));
		return -1;
	}

	/* シリアルポートを読み書き可能な非制御端末としてオープン */
	fd = open(argv[1], O_RDWR);
	if (fd < 0) {
		perror("open");
		return -1;
	}

	/* 終了シグナルに対してハンドラを設定 */
	ret = install_sig_handler();
	if (ret)
		goto err_close;

	/* シリアルポートを設定 */
	ret = setup_serial(fd, &tio);
	if (ret)
		goto err_close;

	/* CRC8テーブル作成 */
	crc8_populate_msb(crc8_table, CRC8_POLYNOMIAL);

	/* ワットチェッカーの初期化を行う */
	init_wattchecker(fd);
	printf("RTC timer set command success\n");

	/* ワットチェッカーの計測コマンド(計測開始コマンド) */
	ret = start_measure(fd);
	if (ret)
		goto err_restore_serial;
	printf("start measure command success\n");

	/* ワットチェッカーのリアルタイム計測データ転送要求コマンド */
	do {
		uint8_t buf[BUF_SIZE];

		ret = request_measure(fd, buf);
		if (ret)
			goto err_stop_measure;

		disp_data_details(buf);

		sleep(REQUEST_INTERVAL);
	} while (!terminated);
	printf("\n");

err_stop_measure:
	/* ワットチェッカーの計測コマンド(計測終了コマンド) */
	ret = stop_measure(fd);
	if (ret)
		goto err_restore_serial;
	printf("stop measure command success\n");

err_restore_serial:
	restore_serial(fd, &tio);

err_close:
	close(fd);

	return ret;
}
